package examples;
import jgame.*;
import java.awt.*;
public class MatrixMiner extends JGEngine {
	public static void main(String [] args) {
		int w=0,h=0;
		if (args.length==2) {
			w = Integer.parseInt(args[0]);
			h = Integer.parseInt(args[1]);
		}
		new MatrixMiner(w,h); 
	}
	public MatrixMiner() { initEngineApplet(19,15,16,16,null,null,null); }
	public MatrixMiner(int x,int y){initEngine(19,15,16,16,null,null,null,x,y);}
	public void initGame() {
		defineImages("examples/matrix_miner.tbl");
		setFrameRate(40,4);
		setGameState("Title");
	}
	int timer, score=0, lives, level;
	String bonus_msg;
	Player player=null;
	public void doFrameTitle() {
		if (getKey(' ')) { // start game
			setGameState("InGame");
			score=0;
			level=0;
			lives=3;
			generateLevel();
		}
	}
	public void doFrameInGame() {
		moveObjects();
		checkCollision(1,8); // player hits goodies
		checkCollision(2+16,1); // enemies and enemy bullets hit player
		checkCollision(2,32); // enemies hit goodies
		checkCollision(4,2); // bullets hit robots
		checkBGCollision(1,4+16); // walls hit bullets
		timer++;
		if (timer%(int)(timer < 30+4*level ? 5 : 55-level*3 ) == 0) {
			switch ((int)random(0,4)) {
				case 0: new Enemy((int)random(1,19,2), -0.8, 0, 1);
				break;
				case 1: new Enemy((int)random(1,19,2), 14.8, 0, -1);
				break;
				case 2: new Enemy(-0.8,(int)random(1,15,2), 1,0);
				break;
				case 3: new Enemy(18.8,(int)random(1,15,2),-1,0);
				break;
			}
		}
		if (countObjects("@goodie",8)==0) {
			setGameState("NextLevel");
			int bonus = countObjects("@goodie",0);
			if (bonus < 20) {
				bonus_msg = "25 X "+bonus+"..."+(25*bonus)+" PTS";
				score += 25*bonus;
			} else {
				bonus_msg = "PERFECT! 1000 PTS";
				score += 1000;
			}
			removeObjects(null,0);
			new JGTimer(100,true) { public void alarm() { // next level
					setGameState("InGame");
					if (level<9) level++;
					generateLevel();
				} };
		}
	}
	public void doFrameDead() { moveObjects(); }
	public void paintFrame() {
		drawImageString(""+score,0,0,"font_map",32,2);
		for (int i=1; i<=lives; i++) drawImage(pfWidth()-16*i,0,"player_d0");
	}
	public void paintFrameTitle() {
		drawImageString("MATRIX MINER",40,75,"font_map",32,2);
		drawImageString("PRESS SPACE",50,140,"font_map",32,2);
	}
	public void paintFrameDead() {
		drawImageString("GOT YOU !",75,75,"font_map",32,2);
	}
	public void paintFrameNextLevel() {
		drawImageString(bonus_msg,0,75,"font_map",32,2);
	}
	public void generateLevel() {
		removeObjects(null,0);
		timer=0;
		player=new Player(160,80);
		clearBG(".");
		for (int y=0; y<pfTilesY(); y++) {
			for (int x=0; x<pfTilesX(); x++) {
				if((!and(x,1) && !and(y,1))
				|| ((and(x,3) && !and(y,1))&&(random(0.0,1.0)>0.8)) ){
					setBGTile(x,y,"#"+level);
				} else
				if (x==0 || x==pfTilesX()-1
				||  y==0 || y==pfTilesY()-1 ) {
					setBGTile(x,y,"*");
				}
			}
		}
		for (int i=0; i<20; i++) {
			new Goodie(16*random(4,15,2), 16*random(4,11,2));
		}
	}
	public class Goodie extends JGObject {
		Point [] prevpos = new Point [4];
		JGObject to_follow=null;
		boolean is_in_tile=true;
		public Goodie(double x,double y) {
			super("@goodie",true,x,y,8,"goodie",0,0,16,16);
			// stick out on all sides so we detect when player passes along
			setBBox(-4,-4,24,24);
		}
		public void hit(JGObject obj) {
			if (colid==32) { // hit with enemy
				remove();
			} else if (is_in_tile) { // player passes along goodie
				int xdir=0,ydir=0;
				if (player.xdir!=0 && player.y > y+8) ydir = -1;
				if (player.xdir!=0 && player.y < y-8) ydir = 1;
				if (player.ydir!=0 && player.x > x+8) xdir = -1;
				if (player.ydir!=0 && player.x < x-8) xdir = 1;
				if (!and(getBGTileCid(getCenterTileTopLeft(),xdir,ydir),1)) {
					clearBBox();
					x += xdir*16;
					y += ydir*16;
					is_in_tile=false;
					score += 5;
				}
			} else { // player collects goodie
				colid=32;
				if (((Player)obj).following!=null)
					((Player)obj).following.to_follow=this;
				((Player)obj).following=this;
				to_follow=obj;
				for (int i=0; i<4; i++) shiftPos(obj.x,obj.y);
				score += 10;
			}
		}
		Point shiftPos(double x,double y) {
			for (int i=0; i<3; i++) prevpos[i] = prevpos[i+1]; 
			prevpos[3] = new Point((int)x,(int)y);
			return prevpos[0];
		}
		public void move() {
			if (to_follow!=null) { // follow the player
				if (!to_follow.isAlive()) remove();
				Point newpos = shiftPos(to_follow.getLastX(),
					to_follow.getLastY() );
				x = newpos.x;
				y = newpos.y;
			}
		}
	}
	public class Player extends JGObject {
		int moving=0;
		Goodie following=null;
		public Player(double x, double y) {
			super("player",false, x,y, 1, "player_u", 0,0); 
		}
		public void move() {
			Point cen = getCenterTileTopLeft();
			if (!isXAligned(2.0) || !isYAligned(2.0)) {
				moving=1;
				startAnim();
			} else {
				snapToGrid(2.1,2.1);
				moving=0;
				stopAnim();
				if (getKey(KeyUp)) {
					if (!and(getBGTileCid(cen,0,-1),3)) {
						setDir(0,-1);
						setAnim("player_u");
					}
					moving=1;
				}
				if (getKey(KeyDown)) {
					if (!and(getBGTileCid(cen,0,1),3)) {
						setDir(0,1);
						setAnim("player_d");
					}
					moving=1;
				}
				if (getKey(KeyLeft)) {
					if (!and(getBGTileCid(cen,-1,0),3)) {
						setDir(-1,0);
						setAnim("player_l");
					}
					moving=1;
				}
				if (getKey(KeyRight)) {
					if (!and(getBGTileCid(cen,1,0),3)) {
						setDir(1,0);
						setAnim("player_r");
					}
					moving=1;
				}
				if (and(getBGTileCid(cen,xdir,ydir),3)) moving=0;
			}
			setSpeed(moving==1  ?  3.0 : 0);
			if (countObjects("bullet",4) < 3) {
				if (getKey('A')) {new Bullet(x,y,-1,0, 4); clearKey('A');}
				else if (getKey('D')) {new Bullet(x,y, 1,0, 4); clearKey('D');}
				else if (getKey('W')) {new Bullet(x,y,0,-1, 4); clearKey('W');}
				else if (getKey('S')) {new Bullet(x,y,0,1,  4); clearKey('S');}
			}
		}
		public void hit(JGObject obj) {
			for (int i=0; i<16; i++) new JGObject("explo",true,x,y,0,"explo",
					random(-2,2),random(-2,2), (int)random(16,64) );
			player.remove();
			setGameState("Dead");
			new JGTimer(100,true) { public void alarm() {
				if ((--lives) <= 0) { // game over
					setGameState("Title");
					removeObjects(null,0);
				} else { // restore game state and continue
					setGameState("InGame");
					removeObjects("Enemy",0);
					removeObjects("bullet",0);
					player=new Player(160,80);
				}
			} };
		}
	}
	public class Bullet extends JGObject {
		public Bullet(double x,double y, int xdir,int ydir, int cid) {
			super("bullet",true,x,y,cid,"bullet", 0,0,-2);
			double speed = and(cid,16) ?  4.0/(3.0 - 0.15*level)  :  4.0;
			setSpeedAbs(xdir*speed, ydir*speed);
		}
		public void hit_bg(int cid) { remove(); }
	}
	public class Enemy extends JGObject {
		int bullettimer=200 - (int)random(0,100);
		public Enemy(double x, double y, int xdir, int ydir) {
			super("Enemy",true,16*x,16*y,2,"enemy");
			setDirSpeed(xdir,ydir, 0.35 + 0.05*level);
		}
		public void move() {
			if (isXAligned(xspeed-0.1) && isYAligned(yspeed-0.1)) {
				snapToGrid(xspeed-0.1, yspeed-0.1);
				// determine preferred direction
				int newxdir,newydir;
				if (Math.abs(x-player.x) > Math.abs(y-player.y)) {
					newxdir = (x > player.x) ? -1 : 1;
					newydir=0;
				} else {
					newxdir=0;
					newydir = (y > player.y) ? -1 : 1;
				}
				Point cen = getCenterTileTopLeft();
				if (and(getBGTileCid(cen,newxdir,newydir),1)) {
					if (newxdir!=0) { // try approaching vertically
						newxdir=0;
						newydir = (y > player.y) ? -1 : 1;
					} else { // or horizontally
						newxdir = (x > player.x) ? -1 : 1;
						newydir=0;
					}
				}
				if (newxdir == -xdir && newydir == -ydir) { // refuse u-turn
					newxdir = xdir;
					newydir = ydir;
				}
				if (and(getBGTileCid(cen,newxdir,newydir),1)) {
					newxdir=0;
					newydir=0;
				}
				setDir(newxdir,newydir);
			}
			if ((xdir!=0 || ydir!=0) && bullettimer<=0) { // shoot
				new Bullet(x,y,xdir,ydir,16);
				bullettimer=440 - 20*level;
			}
			if (bullettimer > 0) bullettimer--;
			if (!isOnScreen(0,0)) remove();
		}
		public void hit(JGObject o) {
			for (int i=0; i<8; i++) new JGObject("explo",true,x,y,0,"explo",
					random(-2,2),random(-2,2), (int)random(8,32) );
			o.remove();
			remove();
			score += 5;
		}
	}
}
