package jgame;
import java.awt.*;
import java.awt.event.*;
import javax.swing.ListCellRenderer;
import java.util.*;
import java.awt.image.ImageObserver;
import java.net.*;

/** JGCanvas is internally used by JGEngine for updating and drawing objects
 * and tiles. You should not need to use this class directly when writing a game
 * application.
 */
class JGCanvas extends Canvas {
	// part of the "official" method of handling keyboard focus
	public boolean isFocusTraversable() { return true; }
	public class AllMouse implements MouseListener, MouseMotionListener {
		public void mouseClicked(MouseEvent e) {
			// part of the "official" method of handling keyboard focus
			requestFocus();
		}
		public void mouseEntered(MouseEvent e) {}
		public void mouseExited(MouseEvent e) {}
		public void mousePressed(MouseEvent e) {}
		public void mouseReleased(MouseEvent e) {}
 		public void mouseDragged(MouseEvent e) {}
		public void mouseMoved(MouseEvent e) {}
	}
	AllMouse mouse_listener = new AllMouse();

	JGEngine engine;
	/** Determines whether repaint will show the game graphics or do
	 * nothing. */
	protected boolean is_initialised=false;
	/** paint interface that is used when the canvas is not initialised (for
	 * displaying status info while starting up, loading files, etc. */
	private ListCellRenderer initpainter=null;
	/** for displaying progress info */
	protected String progress_message=null;

	public JGCanvas (JGEngine engine, int nrtilesx, int nrtilesy,
	int tilex,int tiley,
	int width, int height, URL baseurl,String basepath) {
		super();
		this.engine=engine;
		setSize(width,height);
		this.tilex=tilex;
		this.tiley=tiley;
		this.nrtilesx=nrtilesx;
		this.nrtilesy=nrtilesy;
		this.width_with_pad = width;
		this.height_with_pad = height;
		this.width = ((int)(width / nrtilesx))*nrtilesx;
		this.height = ((int)(height / nrtilesy))*nrtilesy;
		x_scale_fac = width  / (double)(tilex*nrtilesx);
		y_scale_fac = height / (double)(tiley*nrtilesy);
		min_scale_fac = Math.min( x_scale_fac, y_scale_fac );
		setBGDim(tilex,tiley, nrtilesx, nrtilesy, "");
		this.baseurl=baseurl;
		this.basepath=basepath;
		clearKeymap();
		addMouseListener(mouse_listener);
		addMouseMotionListener(mouse_listener);
	}

	private void clearKeymap() {
		for (int i=0; i<256; i++) keymap[i]=false;
	}

	protected void setInitialised() {
		is_initialised=true; 
		initpainter=null;
	}
	protected void setInitPainter(ListCellRenderer painter) {
		initpainter=painter;
	}
	protected void setProgressMessage(String msg) {
		progress_message=msg;
		if (!is_initialised && initpainter!=null) repaint();
	}


	protected Point scalePos(double x, double y) {
		return new Point(
			(int)Math.floor(((double)width/(double)(tilex*nrtilesx)) * x),
			(int)Math.floor(((double)height/(double)(tiley*nrtilesy)) * y) );
	}

	/** returns null if r is null */
	protected Rectangle scalePos(Rectangle r) {
		if (r==null) return null;
		Point topleft = scalePos(r.x, r.y);
		Point botright = scalePos(r.x+r.width, r.y+r.height);
		return new Rectangle(topleft.x, topleft.y,
			botright.x-topleft.x - 1, botright.y-topleft.y - 1);
	}

	public ImageUtil imageutil = new ImageUtil(this);
	//public MediaTracker mediatracker = new MediaTracker(this);
	//int next_id=0;

	/** Strings -&gt; Images, original size,
	* nonexistence means there is no image */
	protected Hashtable images_orig = new Hashtable();
	/** Strings -&gt; Images, screen size, nonexistence indicates image
	* is not cached and needs to be generated from images_orig */
	protected Hashtable images = new Hashtable();
	protected Hashtable images_transp = new Hashtable(); 
		/* Integers -> Objects, existence indicates transparency */
	protected Hashtable images_tile = new Hashtable(); /* Integers -> Strings */
	protected Hashtable images_bbox = new Hashtable(); /* Strings -> Rectangles */
	protected Hashtable images_tilecid = new Hashtable(); /* Integers -> Integers */

	protected Hashtable imagemaps = new Hashtable(); /* Strings->JGImageMaps*/
	class JGImageMap {//implements ImageObserver {
		public Image img;
		//public Image scaled_img=null;
		public int xofs,yofs;
		public int tilex,tiley;
		public int skipx,skipy;
		public JGImageMap (String imgurl, int xofs,int yofs,
		int tilex,int tiley, int skipx,int skipy) {
			URL imageurl=null;
			try {
				imageurl=new URL(baseurl,basepath+imgurl);
			} catch (MalformedURLException e) {
				e.printStackTrace();
			}
			setProgressMessage(imageurl.toString());
			img = imageutil.loadImage(imageurl);
			this.xofs=xofs;
			this.yofs=yofs;
			this.tilex=tilex;
			this.tiley=tiley;
			this.skipx=skipx;
			this.skipy=skipy;
		}
		//public int getHeight() { return img.getHeight(this); }
		/** returns null when image was not loaded */
		public Point getImageCoord(int imgnr) {
			if (img==null) return null;
			Dimension size = imageutil.getSize(img);
			int imgs_per_line = (size.width - xofs + skipx) / (tilex+skipx);
			int ynr = imgnr / imgs_per_line;
			int xnr = imgnr % imgs_per_line;
			return new Point(
				xofs + xnr*(tilex+skipx),
				yofs + ynr*(tiley+skipy) );
		}
		//public Image getScaledImage() {
		//	if (scaled_img!=null) return scaled_img;
		//	Dimension size = imageutil.getSize(img);
		//	Point scaledpos = scalePos(size.width,size.height);
		//	scaled_img = imageutil.scale(img,scaledpos.x,scaledpos.y);
		//	return scaled_img;
		//}
		//public boolean
		//imageUpdate(Image img,int infoflags,int x,int y,int width,int height){
		//	return true;
		//}
	}

	protected Hashtable objects=new Hashtable();      /* String->JGObject */
	protected Vector obj_to_remove = new Vector(20,40); /* String */
	protected Vector obj_spec_to_remove = new Vector(20,40); /* (String,Int) */
	protected Vector obj_to_add = new Vector(20,40); /* JGObject */
	//protected boolean clear_objects=false;

	protected int nrtilesx, nrtilesy;
	protected int width_with_pad,height_with_pad;
	protected int width,height;
	/* min (scalex,scaley); is 1.0 until width, height is defined */
	protected double x_scale_fac=1.0, y_scale_fac=1.0, min_scale_fac=1.0;
	protected URL baseurl;
	protected String basepath;

	Image background=null,buffer=null;
	
	protected int xofs=0,yofs=0;
	protected int tilex=32,tiley=32;
	protected int virtualx,virtualy;
	/** Image name (not tile name) of imag to use behind transparent tiles.
	* Null indicates: use solid fill */
	protected String bg_image=null;
	protected Dimension bg_image_tiles=null;
	protected int [] [] tilemap=null;
	protected int [] [] tilecidmap=null;

	//Vector keybuf = new Vector(50,50);
	protected boolean [] keymap = new boolean [256];
	protected boolean do_repaint_all=false;

	/* ImageObserver method */

//	public boolean
//	imageUpdate(Image img, int infoflags, int x,int y, int width,int height) {
//		/* Apparently, Toolkit.getImage() does not always load the image
//		 * completely, or causes some other images to unload, so we need to
//		 * check for images that are completed afterwards */
//		if (infoflags == ImageObserver.ALLBITS) repaint();
//		//int width = img.getWidth(i);
//		//int height = img.getHeight(i);
//		do_repaint_all=true;
//		return true;
//	}


	protected boolean existsImage(String imgname) {
		return images.containsKey(imgname);
	}

	protected boolean existsTileImage(int tileid) {
		Integer tileid_obj = new Integer(tileid);
		if (!images_tile.containsKey(tileid_obj)) return false;
		return images_orig.containsKey(images_tile.get(tileid_obj));
	}

	/** returns null if image not loaded or null.  The image may still have
	 * bounding box and other properties. */
	protected Image getImage(String imgname) {
		Image img = (Image)images.get(imgname);
		if (img==null) {
			img = (Image)images_orig.get(imgname);
			if (img==null) return null;
			//int width = img.getWidth(this);
			//int height = img.getHeight(this);
			Dimension size = imageutil.getSize(img);
			if (width>0 && height>0) {
				Point scaledpos = scalePos(size.width,size.height);
				//System.out.println("Yes!"+scaledpos);
				img = imageutil.scale(img, scaledpos.x,scaledpos.y);
				img = imageutil.toBuffered(img);
				images.put(imgname,img);
			} else {
				System.err.println("Width, height < 0 !");
			}
		}
		return img;
	}

	protected Image getSubImage(String mapname,int imgnr) {
		JGImageMap imgmap = (JGImageMap)imagemaps.get(mapname);
		Point subcoord = imgmap.getImageCoord(imgnr);
		if (subcoord!=null) {
			return imageutil.crop(imgmap.img,subcoord.x,subcoord.y,
					imgmap.tilex,imgmap.tiley);
		} else {
			return null;
		}
//		Image subimg = createImage(imgmap.tilex,imgmap.tiley);
//		Graphics g = subimg.getGraphics();
//		g.setColor(new Color(250,0,0,0));
//		g.fillRect(0,0,imgmap.tilex,imgmap.tiley);
//		g.drawImage(imgmap.img,
//			-subcoord.x,-subcoord.y,this);
//		//System.out.println("coord "+subcoord);
//		//try { Thread.sleep(2000); } catch (InterruptedException e) {}
//		return subimg;
	}

	protected void drawImage(Graphics g,int x,int y,String imgname) {
		if (imgname==null) return;
		Point scaled = scalePos(x,y);
		Image img = getImage(imgname);
		if (img!=null) g.drawImage(img,scaled.x,scaled.y,this);
	}

	protected void drawObject(Graphics g, JGObject o) {
		o.prepareForFrame();
		drawImage(g,(int)o.x,(int)o.y,o.imgname);
		if ((JGEngine.debugflags&1)!=0) {
			g.setColor(engine.fg_color);
			Rectangle bbox = scalePos(o.getBBox());
			g.drawRect(bbox.x,bbox.y,bbox.width,bbox.height);
			bbox = scalePos(o.getTileBBox());
			g.drawRect(bbox.x,bbox.y,bbox.width,bbox.height);
			// XXX colours should be configurable
			g.setColor(Color.green);
			bbox = getTileSpan(o.getTileBBox());
			bbox.x *= tilex;
			bbox.y *= tiley;
			bbox.width *= tilex;
			bbox.height *= tiley;
			bbox = scalePos(bbox);
			g.drawRect(bbox.x,bbox.y,bbox.width,bbox.height);
			g.setColor(Color.magenta);
			bbox = o.getCenterTiles();
			bbox.x *= tilex;
			bbox.y *= tiley;
			bbox.width *= tilex;
			bbox.height *= tiley;
			bbox = scalePos(bbox);
			g.drawRect(bbox.x+2,bbox.y+2,bbox.width-2,bbox.height-2);
		}
		o.frameFinished();
	}

	/** 
	* @param imgfile  filespec relative to baseurl; "null" means no file
	*/
	protected void defineImage(String name, String tilename, int collisionid,
	String imgfile, boolean is_transparent, String img_op,
	int top,int left, int width,int height) {
		try {
			Image img=null;
			if (!imgfile.equals("null")) {
				//System.out.println(baseurl+"#######"+basepath+"####"+imgfile);
				URL imageurl=new URL(baseurl,basepath+imgfile);
				setProgressMessage(imageurl.toString());
				img = imageutil.loadImage(imageurl);
			}
			defineImage(name,tilename, collisionid, img, is_transparent,
				img_op, top,left, width,height);
		} catch (MalformedURLException e) {
			e.printStackTrace();
		}
	}

	/** 
	* @param imgfile  filespec relative to baseurl
	*/
	protected void defineImage(String name, String tilename, int collisionid,
	Image img, boolean is_transparent, String img_op,
	int top,int left, int width,int height) {
		if (img!=null) {
			/* do image ops */
			img_op = img_op.toLowerCase();
			boolean flipx = img_op.indexOf("x") >= 0;
			boolean flipy = img_op.indexOf("y") >= 0;
			boolean rot90  = img_op.indexOf("r") >= 0;
			boolean rot180 = img_op.indexOf("u") >= 0;
			boolean rot270 = img_op.indexOf("l") >= 0;
			//System.out.println("img_op "+img_op+ " "+flipx+" "+flipy);
			if (flipx || flipy) img = imageutil.flip(img,flipx,flipy);
			if (rot90) { img = imageutil.rotate(img,90); }
			else if (rot180) { img = imageutil.rotate(img,180); }
			else if (rot270) { img = imageutil.rotate(img,270); }
			images_orig.put(name,img);
		}
		Integer tileid = new Integer(tilestrToInt(tilename));
		if (is_transparent) images_transp.put(tileid, new Object());
		images_tile.put(tileid, name);
		images_tilecid.put(tileid, new Integer(collisionid));
		images_bbox.put(name,new Rectangle(top,left,width,height));
		//getImage(name); /* pre-load scaled image */
	}

	protected void defineImageMap(String mapname, String imgurl,
		int xofs,int yofs, int tilex,int tiley, int skipx,int skipy) {
		imagemaps.put(mapname, new JGImageMap (imgurl, xofs,yofs,
			tilex,tiley, skipx,skipy) );
	}

	protected Rectangle getImageBBox(String imgname) {
		return (Rectangle)images_bbox.get(imgname);
	}


	/** null=turn off bg image */
	protected void setBGImage(String bgimg) {
		bg_image=bgimg;
		bg_image_tiles = imageutil.getSize((Image)images_orig.get(bgimg));
		bg_image_tiles.width /= tilex;
		bg_image_tiles.height /= tiley;
	}



	protected void setBGDim(int tilex,int tiley, int virtualx,int virtualy,
	String filltile) {
		this.tilex=tilex;
		this.tiley=tiley;
		this.virtualx=virtualx;
		this.virtualy=virtualy;
		allocTileMap();
		clearBG(filltile);
	}

	private void allocTileMap() {
		tilemap = new int [virtualx][virtualy];
		tilecidmap = new int [virtualx][virtualy];
	}

	protected void clearBG(String filltile) {
		//this.filltile=filltile;
		//if (existsTileImage(tilestrToInt(filltile))) {
		//	filltileid=tilestrToInt(filltile);
		//} else {
		//	filltileid=0;
		//}
		for (int y=0; y<virtualy; y++) {
			for (int x=0; x<virtualx; x++) {
				setBGTile(x,y,filltile);
			}
		}
	}

	protected void setBGTile(int x,int y,String tilestr) {
		if (x<0 || y<0 || x>=virtualx || y>=virtualy) return;
		int tileid = tilestrToInt(tilestr);
		tilemap[x][y] = tileid;
		tilecidmap[x][y] = tilestrToCid(tilestr);
		drawBGTile(x,y,tileid);
	}


	/** xi,yi must be within the tilemap */
	protected void drawBGTile(int xi,int yi,int tileid) {
		if (background != null) {
			int x = xi * tilex;
			int y = yi * tiley;
			Graphics bgg = background.getGraphics();
			// define background behind tile in case the tile is null or
			// transparent.
			boolean tile_defined = !( tileid==0 || !existsTileImage(tileid) );
			if (!tile_defined || images_transp.containsKey(new Integer(tileid))) {
				Point scp = scalePos(x,y);
				Point scd = scalePos(tilex,tiley);
				if (bg_image==null) {
					bgg.setColor(engine.bg_color);
					bgg.fillRect(scp.x,scp.y,scd.x,scd.y);
				} else {
					int xtile = xi % bg_image_tiles.width;
					int ytile = yi % bg_image_tiles.height;
					bgg.drawImage(getImage(bg_image),
						scp.x, scp.y, scp.x+scd.x, scp.y+scd.y,
						xtile*scd.x, ytile*scd.y, 
						(xtile+1)*scd.x, (ytile+1)*scd.y,
						engine.bg_color,
						null);
					//drawImage(bgg,x,y,bg_image);
				}
			}
			if (tile_defined) {
				drawImage(bgg,x-xofs,y-yofs,
					(String)images_tile.get(new Integer(tileid)) );
			}
		}
		//System.out.println("Drawn tile"+tileid);
	}

	protected void repaintBG() {
		for (int x=0; x<virtualx; x++)
			for (int y=0; y<virtualy; y++)
				drawBGTile(x,y,tilemap[x][y]);
		do_repaint_all=false;
	}

	/* object methods */

	/* only to be called by package */

	/** Add new object, will become active next frame. */
	protected void markAddObject(JGObject obj) {
		obj_to_add.add(obj);
	}

	/** Add new object now. */
	protected void addObject(JGObject obj) {
		objects.put(obj.name,obj);
	}

	/** Mark object for removal. */
	protected void markRemoveObject(String index) {
		obj_to_remove.add(index);
	}

	/** Mark object for removal. */
	protected void markRemoveObject(JGObject obj) {
		obj_to_remove.add(obj.name);
	}

	/** Actually remove object now */
	protected void removeObject(JGObject obj) {
		obj.removeDone();
		objects.remove(obj.name);
	}

	/** Mark all objects with given spec for removal. */
	protected void markRemoveObjects(String prefix,int cidmask) {
		obj_spec_to_remove.add(prefix);
		obj_spec_to_remove.add(new Integer(cidmask));
	}
	/** Actually remove objects with given spec, including those in obj_to_add
	 * list. */
	protected void removeObjects(String prefix,int cidmask) {
		for (Enumeration e=objects.keys(); e.hasMoreElements();) {
			String name = (String) e.nextElement();
			if (prefix==null || name.startsWith(prefix)) {
				JGObject o = (JGObject) getObject(name);
				if (cidmask==0 || (o.colid&cidmask)!=0) {
					removeObject(o);
				}
			}
		}
		for (int i=obj_to_add.size()-1; i>=0; i--) {
			JGObject o = (JGObject) obj_to_add.get(i);
			if (prefix==null || o.name.startsWith(prefix)) {
				if (cidmask==0 || (o.colid&cidmask)!=0) {
					obj_to_add.remove(i);
				}
			}
		}
	}


	/** Remove objects marked for removal. */
	protected void flushRemoveList() {
		//if (clear_objects) objects.clear();
		//clear_objects=false;
		for (Enumeration e=obj_to_remove.elements(); e.hasMoreElements();) {
			String name = (String)e.nextElement();
			JGObject o = (JGObject)objects.get(name);
			if (o!=null) { // object might have been removed already
				removeObject(o);
			}
		}
		for (Enumeration e=obj_spec_to_remove.elements();e.hasMoreElements();){
			String prefix = (String) e.nextElement();
			int cid = ((Integer)e.nextElement()).intValue();
			removeObjects(prefix,cid);
		}
		obj_to_remove.clear();
		obj_spec_to_remove.clear();
	}

	/** Add objects marked for addition.
	*/
	protected void flushAddList() {
		for (Enumeration e=obj_to_add.elements(); e.hasMoreElements();) {
			addObject((JGObject)e.nextElement());
		}
		obj_to_add.clear();
	}

	/** Remove objects marked for addition before they can be added.
	*/
	protected void clearAddList() {
		for (Enumeration e=obj_to_add.elements(); e.hasMoreElements();) {
			JGObject o = (JGObject)e.nextElement();
			o.removeDone(); // be sure to mark the object as removed
		}
		obj_to_add.clear();
	}


	/* public */

	/** Get object if it exists, null if not.
	*/
	protected boolean existsObject(String index) {
		return objects.containsKey(index);
	}

	/** Get object if it exists.
	*/
	protected JGObject getObject(String index) {
		return (JGObject)objects.get(index);
	}

	/** Remove all objects.  All objects are marked for removal, the add
	* list is cleared.  */
	protected void clearObjects() {
		for (Enumeration e=objects.keys(); e.hasMoreElements(); ) {
			markRemoveObject((String)e.nextElement());
		}
		clearAddList();
		//clear_objects=true;
	}


	/** Move only those objects with given specifications.
	* @param cidmask collision id mask, 0 means ignore
	* @param prefix  ID prefix, null means ignore  */
	protected void moveObjects(String prefix, int cidmask) {
		for (Enumeration e=objects.keys(); e.hasMoreElements(); ) {
			String name = (String) e.nextElement();
			if (prefix!=null && !name.startsWith(prefix)) continue;
			JGObject o  = (JGObject)objects.get(name);
			if (cidmask!=0 && (o.colid&cidmask)==0) continue;
			o.move();
			o.x += o.xdir*o.xspeed;
			o.y += o.ydir*o.yspeed;
			if (o.expiry==0) o.remove();
			if (o.expiry > 0) o.expiry--;
			if (o.expiry==-2 && !o.isOnScreen(8,8)) o.remove();
		}
		flushRemoveList();
	}


	protected int getBGTileCid(int xidx,int yidx) {
		if (xidx<0 || yidx<0 || xidx>=virtualx || yidx>=virtualy) return 0;
		return tilecidmap[xidx][yidx];
	}

	protected String getBGTileString(int xidx,int yidx) {
		if (xidx<0 || yidx<0 || xidx>=virtualx || yidx>=virtualy) return null;
		return tileidToString(tilemap[xidx][yidx]);
	}

	protected int getBGTileCid(Rectangle tiler) {
		int cid=0;
		for (int x=tiler.x; x<tiler.x+tiler.width; x++)
			for (int y=tiler.y; y<tiler.y+tiler.height; y++)
				cid |= getBGTileCid(x,y);
		return cid;
	}

	private int tilestrToCid(String tilestr) {
		if (tilestr==null || tilestr.length()==0) return 0;
		Integer tileid = (Integer)
			images_tilecid.get( new Integer(tilestrToInt(tilestr)) );
		if (tileid==null) {
			System.err.println("Warning: unknown tile '"+tilestr+"'");
			return 0;
		}
		return tileid.intValue();
	}

	/** null or empty string -&gt; 0 */
	private int tilestrToInt(String tilestr) {
		if (tilestr==null) return 0;
		switch (tilestr.length()) {
			case 0: return 0;
			case 1: return (int)tilestr.charAt(0);
			case 2: return (int)tilestr.charAt(0)
			             + (int)tilestr.charAt(1)*256;
			case 3: return (int)tilestr.charAt(0)
			             + (int)tilestr.charAt(1)*256
			             + (int)tilestr.charAt(2)*256*256;
			case 4: return (int)tilestr.charAt(0)
			             + (int)tilestr.charAt(1)*256
			             + (int)tilestr.charAt(2)*256*256
			             + (int)tilestr.charAt(3)*256*256*256;
			default:
				System.out.println("String '"+tilestr+" has wrong size!");
				return 0;
		}
	}

	/** tileid==0 -&gt; empty string */
	private String tileidToString(int tileid) {
		if (tileid==0) return "";
		StringBuffer tilestr = new StringBuffer(""+(char)(tileid&255));
		if (tileid >= 0x100) tilestr.append( (char)((tileid/0x100)&255));
		if (tileid >= 0x10000) tilestr.append( (char)((tileid/0x10000)&255));
		if (tileid>=0x1000000) tilestr.append( (char)((tileid/0x1000000)&255));
		return tilestr.toString();
	}

	/** Get tile position range of all tiles overlapping given rectangle.
	* Returns null is rectangle is null.
	* @param r   rectangle in tile coordinates, null is none  */
	protected Rectangle getTileSpan(Rectangle r) {
		if (r==null) return null;
		Rectangle tiler = new Rectangle(r);
		tiler.x /= tilex;
		tiler.y /= tiley;
		tiler.width  = 1 - tiler.x + (r.x + r.width  - 1) / tilex;
		tiler.height = 1 - tiler.y + (r.y + r.height - 1) / tiley;
		return tiler;
	}


	/** Calls all colliders of objects that match dstid that collide with
	* objects that match srcid.
	*/
	protected void checkCollision(int srcid,int dstid) {
		/* get all matching objects */
		Vector srcobj=new Vector(50,50);
		Vector dstobj=new Vector(50,50);
		for (Enumeration e=objects.keys(); e.hasMoreElements(); ) {
			String okey = (String)  e.nextElement();
			JGObject o  = (JGObject)objects.get(okey);
			Rectangle r = o.getBBox();
			if (r!=null) {
				if ((o.colid & srcid) != 0) {
					srcobj.add(o);
				}
				if ((o.colid & dstid) != 0) {
					dstobj.add(o);
				}
			}
		}
		/* check collision */
		for (Enumeration e=srcobj.elements(); e.hasMoreElements(); ) {
			JGObject srco = (JGObject) e.nextElement();
			Rectangle sr = srco.getBBox();
			if (sr==null) continue;
			for (Enumeration f=dstobj.elements(); f.hasMoreElements(); ) {
				JGObject dsto = (JGObject) f.nextElement();
				if (dsto == srco) continue;
				Rectangle dr = dsto.getBBox();
				if (dr==null) continue;
				if (sr.intersects(dr)) {
					dsto.hit(srco);
				}
			}
		}
		flushRemoveList();
	}

	/** Check collision of tiles within given rectangle, return the OR of all
	* cids found.
	*/
	protected int checkBGCollision(Rectangle r) {
		return getBGTileCid(getTileSpan(r));
	}

	/** Calls all bg colliders of objects that match objid that collide with
	* tiles that match tileid.
	*/
	protected void checkBGCollision(int tileid,int objid) {
		/* get all matching objects */
		Vector obj=new Vector(50,50);
		for (Enumeration e=objects.elements(); e.hasMoreElements(); ) {
			JGObject o = (JGObject) e.nextElement();
			Rectangle r = o.getTileBBox();
			if (r!=null) if ((o.colid & objid) != 0) {
				obj.add(o);
			}
		}
		/* check collision */
		for (Enumeration e=obj.elements(); e.hasMoreElements(); ) {
			JGObject o = (JGObject) e.nextElement();
			Rectangle r = o.getTileBBox();
			if (r==null) continue;
			int cid=checkBGCollision(r);
			if ((cid & tileid) != 0) {
				Rectangle tiler = getTileSpan(r);
				o.hit_bg(cid);
				o.hit_bg(cid,tiler.x,tiler.y,tiler.width,tiler.height);
			}
		}
		flushRemoveList();
	}


	/* canvas methods */

	/** Don't call directly. Use repaint().
	*/
	public void update(Graphics g) { paint(g); }

	/** Don't call directly. Use repaint().
	*/
	public synchronized void paint(Graphics g) {
		if (!is_initialised) {
			if (initpainter!=null) {
				if (buffer==null) {
					buffer=createImage(width,height);
				}
				initpainter.getListCellRendererComponent(null,
						buffer.getGraphics(),0,false,false);
				g.drawImage(buffer,0,0,this);
			}
			return;
		}
		if (background==null) {
			background=createImage(width,height);
			do_repaint_all=true;
		}
		if (do_repaint_all) {
			repaintBG();
		}
		if (buffer==null) {
			buffer=createImage(width,height);
		}
		if (buffer!=null && background!=null) {
			/* clear buffer */
			Graphics bufg = buffer.getGraphics();
			//bufg.setColor(getBackground());
			bufg.drawImage(background,0,0,this);
			//Color defaultcolour=g.getColor();
			/* sort objects */
			ArrayList sortedkeys = new ArrayList(objects.keySet());
			// XXX this routine occasionally gets null pointer errors on some
			// machines
			Collections.sort(sortedkeys);
			for (Iterator i=sortedkeys.iterator(); i.hasNext(); ) {
				JGObject o = (JGObject) objects.get(i.next());
				if (o==null) {
					// it is not known why this happens. we made sure that the
					// hashtable is never updated while being enumerated.
					//System.out.println("Warning: key with null value!!");
				} else {
					drawObject(bufg,o);
				}
			}
			/* draw status */
			if (bufg!=null) engine.paintFrame(bufg);
			/* draw buffer */
			g.drawImage(buffer,0,0,this);
			//g.setColor(defaultcolour);
			Toolkit.getDefaultToolkit().sync();
		}
	}

}

